var __defProp = Object.defineProperty;
var __defProps = Object.defineProperties;
var __getOwnPropDescs = Object.getOwnPropertyDescriptors;
var __getOwnPropSymbols = Object.getOwnPropertySymbols;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __propIsEnum = Object.prototype.propertyIsEnumerable;
var __defNormalProp = (obj, key, value) => key in obj ? __defProp(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __spreadValues = (a, b) => {
  for (var prop in b || (b = {}))
    if (__hasOwnProp.call(b, prop))
      __defNormalProp(a, prop, b[prop]);
  if (__getOwnPropSymbols)
    for (var prop of __getOwnPropSymbols(b)) {
      if (__propIsEnum.call(b, prop))
        __defNormalProp(a, prop, b[prop]);
    }
  return a;
};
var __spreadProps = (a, b) => __defProps(a, __getOwnPropDescs(b));
var __objRest = (source, exclude) => {
  var target = {};
  for (var prop in source)
    if (__hasOwnProp.call(source, prop) && exclude.indexOf(prop) < 0)
      target[prop] = source[prop];
  if (source != null && __getOwnPropSymbols)
    for (var prop of __getOwnPropSymbols(source)) {
      if (exclude.indexOf(prop) < 0 && __propIsEnum.call(source, prop))
        target[prop] = source[prop];
    }
  return target;
};

// src/utils/requestScreen.ts
var screenElement;
function requestScreen(customScreen, { zIndex } = {}) {
  if (!customScreen) {
    return { width: window.innerWidth, height: window.innerHeight, x: 0, y: 0 };
  }
  if (customScreen instanceof DOMRect) {
    return { width: customScreen.width, height: customScreen.height, x: customScreen.x, y: customScreen.y };
  }
  const screenEl = customScreen instanceof HTMLElement ? customScreen : document.createElement("div");
  if (typeof customScreen === "object" && customScreen !== null && !(customScreen instanceof HTMLElement)) {
    Object.assign(screenEl.style, customScreen);
  }
  screenEl.style.position = "fixed";
  screenEl.style.display = "flex";
  screenEl.style.alignItems = "center";
  screenEl.style.justifyContent = "center";
  screenEl.style.top = screenEl.style.top || "0px";
  screenEl.style.left = screenEl.style.left || "0px";
  screenEl.style.width = screenEl.style.width || "0px";
  screenEl.style.height = screenEl.style.height || "0px";
  if (zIndex !== void 0) {
    screenEl.style.zIndex = zIndex.toString();
  }
  screenElement = screenEl;
  document.body.appendChild(screenEl);
  const { width, height, x, y } = screenEl.getBoundingClientRect();
  return { width, height, x, y };
}
function clearScreen() {
  screenElement == null ? void 0 : screenElement.remove();
  screenElement = void 0;
}

// src/utils/requestCursor.ts
var cursorElement;
function requestCursor(customCursor, { zIndex } = {}) {
  const cursorEl = document.createElement("div");
  const cursorChildWrapper = document.createElement("div");
  cursorEl.appendChild(cursorChildWrapper);
  if (customCursor instanceof HTMLElement) {
    cursorChildWrapper.appendChild(customCursor);
  } else if (typeof customCursor === "string") {
    cursorChildWrapper.textContent = customCursor;
  } else if (typeof customCursor === "object" && customCursor !== null) {
    Object.assign(cursorChildWrapper.style, customCursor);
  }
  cursorEl.style.position = "fixed";
  cursorEl.style.top = "0px";
  cursorEl.style.left = "0px";
  cursorChildWrapper.style.transform = "translate(-50%, -50%)";
  if (zIndex !== void 0) {
    cursorEl.style.zIndex = zIndex.toString();
  }
  cursorElement = cursorEl;
  document.body.appendChild(cursorEl);
  return cursorEl;
}
function clearCursor() {
  cursorElement == null ? void 0 : cursorElement.remove();
  cursorElement = void 0;
}

// src/pointer-lock-movement.ts
var HISTORY_LENGTH = 8;
var CONFIDENCE_THRESHOLD = 2;
function isSupportPointerLock() {
  return "pointerLockElement" in document || "mozPointerLockElement" in document || "webkitPointerLockElement" in document;
}
function assertSupportPointerLock() {
  if (isSupportPointerLock()) {
    return;
  }
  throw new Error("Your browser does not support pointer lock");
}
var pointerLockMovement = (element, _a = {}) => {
  var _b = _a, {
    loopBehavior = "loop",
    trigger = "drag",
    zIndex = 99999
  } = _b, customOption = __objRest(_b, [
    "loopBehavior",
    "trigger",
    "zIndex"
  ]);
  const options = __spreadProps(__spreadValues({}, customOption), { loopBehavior, trigger, zIndex });
  function requestPointerLock() {
    var _a2, _b2;
    ((_b2 = (_a2 = element.requestPointerLock) != null ? _a2 : element.mozRequestPointerLock) != null ? _b2 : element.webkitRequestPointerLock).call(element);
  }
  function exitPointerLock() {
    var _a2, _b2;
    ((_b2 = (_a2 = document.exitPointerLock) != null ? _a2 : document.mozExitPointerLock) != null ? _b2 : document.webkitExitPointerLock).call(document);
  }
  function isLocked() {
    return element === document.pointerLockElement || element === document.mozPointerLockElement || element === document.webkitPointerLockElement;
  }
  const move = (context, effect) => (payload) => {
    let { movementX, movementY } = payload;
    const historyX = [...context.historyX, movementX].slice(-HISTORY_LENGTH);
    const historyY = [...context.historyY, movementY].slice(-HISTORY_LENGTH);
    const shouldValidate = historyX.length === HISTORY_LENGTH && historyY.length === HISTORY_LENGTH;
    if (shouldValidate) {
      const averageX = historyX.reduce((sum, value) => sum + value, 0) / historyX.length;
      const averageY = historyY.reduce((sum, value) => sum + value, 0) / historyY.length;
      const standardDeviationX = Math.sqrt(historyX.reduce((sum, value) => sum + Math.pow(value - averageX, 2), 0) / historyX.length);
      const standardDeviationY = Math.sqrt(historyY.reduce((sum, value) => sum + Math.pow(value - averageY, 2), 0) / historyY.length);
      if (standardDeviationX > 0) {
        const zScoreX = Math.abs(movementX - averageX) / standardDeviationX;
        if (zScoreX > CONFIDENCE_THRESHOLD) {
          movementX = averageX;
        }
      }
      if (standardDeviationY > 0) {
        const zScoreY = Math.abs(movementY - averageY) / standardDeviationY;
        if (zScoreY > CONFIDENCE_THRESHOLD) {
          movementY = averageY;
        }
      }
    }
    const contextPatch = {
      event: payload,
      movementX,
      movementY,
      x: context.x + context.movementX,
      y: context.y + context.movementY,
      historyX,
      historyY,
      status: "moving"
    };
    if (options.loopBehavior === "loop") {
      if (contextPatch.x > context.maxWidth) {
        contextPatch.x -= context.maxWidth;
      } else if (contextPatch.x < 0) {
        contextPatch.x += context.maxWidth;
      }
      if (contextPatch.y > context.maxHeight) {
        contextPatch.y -= context.maxHeight;
      } else if (contextPatch.y < 0) {
        contextPatch.y += context.maxHeight;
      }
    } else if (options.loopBehavior === "stop") {
      if (contextPatch.x > context.maxWidth) {
        contextPatch.x = context.maxWidth;
        context.status = "stopped";
      } else if (contextPatch.x < 0) {
        contextPatch.x = 0;
        contextPatch.status = "stopped";
      }
      if (contextPatch.y > context.maxHeight) {
        contextPatch.y = context.maxHeight;
        context.status = "stopped";
      } else if (contextPatch.y < 0) {
        contextPatch.y = 0;
        contextPatch.status = "stopped";
      }
    }
    const newContext = __spreadValues(__spreadValues({}, context), contextPatch);
    effect(newContext);
    if (newContext.event.defaultPrevented) {
      return move(context, effect);
    }
    return move(newContext, effect);
  };
  function startup() {
    let nextFn;
    const localState = {
      targetOnActiveElement: false,
      isDetecting: false,
      startX: 0,
      startY: 0
    };
    function detectMoveOffset(event) {
      if (!(event instanceof PointerEvent) || !event.buttons || !options.dragOffset) {
        return;
      }
      const offset = Math.sqrt(Math.pow(event.clientX - localState.startX, 2) + Math.pow(event.clientY - localState.startY, 2));
      if (offset < options.dragOffset) {
        return;
      }
      element.removeEventListener("pointermove", detectMoveOffset);
      localState.isDetecting = false;
      active(event);
    }
    function deActive() {
      var _a2;
      exitPointerLock();
      (_a2 = options.onLock) == null ? void 0 : _a2.call(options, false);
      document.removeEventListener("pointermove", handlePointerMove);
      nextFn = void 0;
      clearCursor();
      clearScreen();
    }
    function handlePointerMove(event) {
      if (options.trigger === "drag" && !event.buttons) {
        deActive();
      } else {
        handleContinueMove(event);
      }
    }
    function active(pointerEvent) {
      var _a2;
      const virtualScreen = requestScreen(options.screen, { zIndex: options.zIndex });
      const virtualCursor = requestCursor(options.cursor, { zIndex: options.zIndex });
      nextFn = move(
        {
          event: pointerEvent,
          status: "moving",
          startX: pointerEvent.clientX,
          startY: pointerEvent.clientY,
          movementX: 0,
          movementY: 0,
          historyX: [],
          historyY: [],
          x: pointerEvent.clientX - virtualScreen.x,
          y: pointerEvent.clientY - virtualScreen.y,
          maxWidth: virtualScreen.width,
          maxHeight: virtualScreen.height
        },
        ({ event, status, x, y, startX, startY, movementX, movementY }) => {
          var _a3;
          virtualCursor.style.transform = `translate3D(${x}px, ${y}px, 0px)`;
          (_a3 = options.onMove) == null ? void 0 : _a3.call(
            options,
            event,
            {
              status,
              offsetX: x - startX,
              offsetY: y - startY,
              movementX,
              movementY
            }
          );
        }
      )(pointerEvent);
      document.addEventListener("pointermove", handlePointerMove);
      document.addEventListener("pointerlockchange", function handlePointerLockChange() {
        if (isLocked()) {
          return;
        }
        document.removeEventListener("pointerlockchange", handlePointerLockChange);
        deActive();
      });
      (_a2 = options.onLock) == null ? void 0 : _a2.call(options, true);
      requestPointerLock();
    }
    function handleContinueMove(event) {
      nextFn = nextFn == null ? void 0 : nextFn(event);
    }
    function handleDeActive() {
      if (!isLocked()) {
        return;
      }
      deActive();
    }
    function handleActive(event) {
      if (isLocked()) {
        return;
      }
      active(event);
    }
    function handleDragActive(event) {
      if (!(event instanceof PointerEvent) || event.button !== 0) {
        return;
      }
      if (localState.targetOnActiveElement && options.disableOnActiveElement) {
        return;
      }
      handleActive(event);
    }
    function handleToggleActive(event) {
      if (!(event instanceof PointerEvent) || event.button !== 0) {
        return;
      }
      if (isLocked()) {
        handleDeActive();
      } else {
        handleActive(event);
      }
    }
    function handleDragOffsetActive(event) {
      var _a2;
      if (!(event instanceof PointerEvent) || event.button !== 0 || !options.dragOffset) {
        return;
      }
      if (localState.targetOnActiveElement && options.disableOnActiveElement) {
        return;
      }
      if (isLocked()) {
        return;
      }
      (_a2 = options == null ? void 0 : options.onPrepareLock) == null ? void 0 : _a2.call(options, event);
      localState.isDetecting = true;
      localState.startX = event.clientX;
      localState.startY = event.clientY;
      element.addEventListener("pointermove", detectMoveOffset);
      element.addEventListener("pointerup", function unbindEvent(event2) {
        var _a3;
        if (localState.isDetecting) {
          if (event2 instanceof PointerEvent) {
            (_a3 = options == null ? void 0 : options.onCancelPrepareLock) == null ? void 0 : _a3.call(options, event2);
          }
        }
        element.removeEventListener("pointerup", unbindEvent);
        element.removeEventListener("pointermove", detectMoveOffset);
        localState.isDetecting = false;
      });
    }
    function markElementIsActiveElement(event) {
      if (!(event instanceof PointerEvent) || event.button !== 0 || !options.disableOnActiveElement) {
        return;
      }
      if (event.target === document.activeElement) {
        localState.targetOnActiveElement = true;
      } else {
        localState.targetOnActiveElement = false;
      }
    }
    assertSupportPointerLock();
    if (options.disableOnActiveElement) {
      element.addEventListener("pointerdown", markElementIsActiveElement);
    }
    if (options.trigger === "drag") {
      if (options.dragOffset) {
        element.addEventListener("pointerdown", handleDragOffsetActive);
        document.addEventListener("pointerup", handleDeActive);
        return () => {
          element.removeEventListener("pointermove", handleDragOffsetActive);
          document.removeEventListener("pointerup", handleDeActive);
          element.removeEventListener("pointerdown", markElementIsActiveElement);
        };
      }
      element.addEventListener("pointerdown", handleDragActive);
      document.addEventListener("pointerup", handleDeActive);
      return () => {
        element.removeEventListener("pointerdown", handleDragActive);
        document.removeEventListener("pointerup", handleDeActive);
        element.removeEventListener("pointerdown", markElementIsActiveElement);
      };
    } else {
      element.addEventListener("pointerdown", handleToggleActive);
      return () => {
        element.removeEventListener("pointerdown", handleToggleActive);
        element.removeEventListener("pointerdown", markElementIsActiveElement);
      };
    }
  }
  return startup();
};
export {
  isSupportPointerLock,
  pointerLockMovement
};
